#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "fitsio.h"

int main(int argc, char *argv[])
{
    fitsfile *infptr, *outfptr, *darkfptr, *flatfptr;   /* FITS file pointers defined in fitsio.h */
    int status = 0, ii = 1, iteration = 0, single = 0, hdupos;
    int hdutype, bitpix, bytepix, naxis = 0, nkeys, datatype = 0, anynul;
    int dbitpix, dnaxis=0, fbitpix, fnaxis=0;
    int ddatatype, fdatatype, dbytepix, fbytepix;
    int i;
    long naxes[9] = {1, 1, 1, 1, 1, 1, 1, 1, 1};
    long dnaxes[9] = {1, 1, 1, 1, 1, 1, 1, 1, 1};
    long fnaxes[9] = {1, 1, 1, 1, 1, 1, 1, 1, 1};
    long first, totpix = 0, npix;
    long dtotpix, ftotpix, dnpix, fnpix;
    double *array, bscale = 1.0, bzero = 0.0, nulval = 0.;
    double *darray, *farray;
    char card[81];

    if (argc != 3 && argc !=4 && argc !=5 )
    {
 printf("\n");
 printf("Usage:  imcopy inputImage outputImage[compress] [darkImage] [flat Image]\n");
 printf("\n");
 printf("Copy an input image to an output image, optionally compressing\n");
 printf("or uncompressing the image in the process.\n");
 printf("If the darkImage is specified, it will be subtracted from the image.\n");
 printf("If the flatImage is also specified, it will divide the image after dark subtraction.\n");
      return(0);
    }

    /* Open the input file and create output file */
    fits_open_file(&infptr, argv[1], READONLY, &status);
    if (status != 0) {    
        fits_report_error(stderr, status);
        return(status);
    }

    fits_create_file(&outfptr, argv[2], &status);
    if (status != 0) {    
        fits_report_error(stderr, status);
        return(status);
    }


    /* Open Dark and Flat files */
    if ( argc >3 ){
      fits_open_file(&darkfptr, argv[3], READONLY, &status);
      if (status != 0) {    
        fits_report_error(stderr, status);
        return(status);
      }
    }
    if ( argc >4 ){
      fits_open_file(&flatfptr, argv[4], READONLY, &status);
      if (status != 0) {    
        fits_report_error(stderr, status);
        return(status);
      }
    }

    fits_get_hdu_num(infptr, &hdupos);  /* Get the current HDU position */

    /* Copy only a single HDU if a specific extension was given */ 
    if (hdupos != 1 || strchr(argv[1], '[')) single = 1;

    for (; !status; hdupos++)  /* Main loop through each extension */
    {

      fits_get_hdu_type(infptr, &hdutype, &status);

      if (hdutype == IMAGE_HDU) {

          /* get image dimensions and total number of pixels in image */
          for (ii = 0; ii < 9; ii++)
              naxes[ii] = 1;

          fits_get_img_param(infptr, 9, &bitpix, &naxis, naxes, &status);
          totpix = naxes[0] * naxes[1] * naxes[2] * naxes[3] * naxes[4]
             * naxes[5] * naxes[6] * naxes[7] * naxes[8];
	  /*	  fprintf(stderr," totpix=%d\n",totpix); */

	  if( argc > 3 ){
	    fits_get_img_param(darkfptr, 2, &dbitpix, &dnaxis, dnaxes, &status);
	    dtotpix = naxes[0] * naxes[1];
	    /*	    fprintf(stderr,"dtotpix=%d\n",totpix); */
	  }
	  if( argc > 4 ){
	    fits_get_img_param(flatfptr, 2, &fbitpix, &fnaxis, fnaxes, &status);
	    ftotpix = naxes[0] * naxes[1];
	    /*	  fprintf(stderr,"ftotpix=%d\n",totpix); */
	  }


      }

      if (hdutype != IMAGE_HDU || naxis == 0 || totpix == 0) { 

          /* just copy tables and null images */
          fits_copy_hdu(infptr, outfptr, 0, &status);

      } else {

          /* Explicitly create new image, to support compression */
          fits_create_img(outfptr, bitpix, naxis, naxes, &status);

          /* copy all the user keywords (not the structural keywords) */
          fits_get_hdrspace(infptr, &nkeys, NULL, &status); 

          for (ii = 1; ii <= nkeys; ii++) {
              fits_read_record(infptr, ii, card, &status);
              if (fits_get_keyclass(card) > TYP_CMPRS_KEY)
                  fits_write_record(outfptr, card, &status);
          }

          datatype = TDOUBLE;

          bytepix = abs(bitpix) / 4;
	  if(argc > 3 ) dbytepix = abs(dbitpix) / 4;
          if(argc > 4 ) fbytepix = abs(fbitpix) / 4;

          npix = totpix;
          dnpix = dtotpix;
          fnpix = ftotpix;
          iteration = 0;

          /* try to allocate memory for the entire image */
          /* use double type to force memory alignment */
          array = (double *) calloc(npix, bytepix);

          /* if allocation failed, divide size by 2 and try again */
          while (!array && iteration < 10)  {
              iteration++;
              npix = npix / 2;
              array = (double *) calloc(npix, bytepix);
          }

          if (!array)  {
              printf("Memory allocation error\n");
              return(0);
          }

	  if (argc > 3){
	    /* try to allocate memory for the dark image */
	    darray = (double *) calloc(dnpix, dbytepix);

	    /* if allocation failed, divide size by 2 and try again */
	    while (!darray && iteration < 10)  {
              iteration++;
              dnpix = dnpix / 2;
              darray = (double *) calloc(dnpix, dbytepix);
	    }
	    
	    if (!darray)  {
              printf("Memory allocation error\n");
              return(0);
	    }
	  }

	  if (argc > 4){
	    /* try to allocate memory for the flat image */
	    farray = (double *) calloc(fnpix, fbytepix);
	    
	    /* if allocation failed, divide size by 2 and try again */
	    while (!farray && iteration < 10)  {
              iteration++;
              fnpix = fnpix / 2;
              farray = (double *) calloc(fnpix, fbytepix);
	    }
	    
	    if (!farray)  {
              printf("Memory allocation error\n");
              return(0);
	    }
	  }

          /* turn off any scaling so that we copy the raw pixel values */
          fits_set_bscale(infptr,  bscale, bzero, &status);
          fits_set_bscale(outfptr, bscale, bzero, &status);

          first = 1;
          while (totpix > 0 && !status)
          {
	    /*	    fprintf(stderr,"npix=%d\n",npix); */

             /* read all or part of image then write it back to the output file */
             fits_read_img(infptr, datatype, first, npix, &nulval, array, &anynul, &status);
	     if (status != 0) {    
	       fits_report_error(stderr, status);
	       return(status);
	     }

	     if(argc == 4 ){
	       fits_read_img(darkfptr, datatype, first, npix, &nulval, darray, &anynul, &status);
	       if (status != 0) {    
		 fits_report_error(stderr, status);
		 return(status);
	       }
	       fprintf(stdout,"%s : subtract %s\n",argv[2],argv[3]);
	       for(i=0;i<npix;i++) array[i]-=darray[i]; 
	     }

	     if(argc == 5 ){
	       fits_read_img(darkfptr, datatype, first, npix, &nulval, darray, &anynul, &status);
	       fits_read_img(flatfptr, datatype, first, npix, &nulval, farray, &anynul, &status);
	       if (status != 0) {    
		 fits_report_error(stderr, status);
		 return(status);
	       }
	       fprintf(stdout,"%s : subtract %s and divide by %s \n",argv[2],argv[3],argv[4]);
	       for(i=0;i<npix;i++) array[i]=(array[i]-darray[i])/farray[i]; 
	     }

	     /*	     fprintf(stderr,"Remaining totpix=%d\n",totpix); */

             fits_write_img(outfptr, datatype, first, npix, array, &status);
             totpix = totpix - npix;
             first  = first  + npix;
          }
          free(array);
      }

      if (single) break;  /* quit if only copying a single HDU */
      fits_movrel_hdu(infptr, 1, NULL, &status);  /* try to move to next HDU */
    }

    if (status == END_OF_FILE)  status = 0; /* Reset after normal error */

    fits_close_file(outfptr,  &status);
    fits_close_file(infptr, &status);

    /* if error occurred, print out error message */
    if (status)
       fits_report_error(stderr, status);
    return(status);
}
